@@ -2,7 +2,7 @@
    This file is part of LibTMCG.
 
  Copyright (C) 2005, 2006, 2007, 
-               2015, 2016, 2017  Heiko Stamer <HeikoStamer@gmx.net>
+               2015, 2016, 2017, 2018  Heiko Stamer <HeikoStamer@gmx.net>
 
    LibTMCG is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -460,6 +460,7 @@ yTIVNwGjZ3pM73jsUA2RxCMfjHntG81euIBZgn8evIJRNvimC8aRh7ITAuU3soQSdQiIld2d\
 	}
 	stop_clock();
 	std::cout << elapsed_time() << std::endl;
+
 	// check whether spowm() is slower than powm()
 	assert((compare_elapsed_time_saved(0) < 0));
 	size_t bad_cnt = 0; 
@@ -527,11 +528,11 @@ yTIVNwGjZ3pM73jsUA2RxCMfjHntG81euIBZgn8evIJRNvimC8aRh7ITAuU3soQSdQiIld2d\
 	}
 	delete [] dig1, delete [] dig2;
 	std::cout << "mpz_shash()" << std::endl;
-	mpz_set_str(bar, "RccLJ5STdkwhAE0HMcrKlxj9ivWHS5LfU4FmAk4hYBF", TMCG_MPZ_IO_BASE);
+	mpz_set_str(bar, "KlGMM70snzDNYXwufIkOkLZzs91jPNPe7QYT9Agfeg0", TMCG_MPZ_IO_BASE);
 	mpz_set_ui(foo2, 23L), mpz_set_ui(bar2, 42L);
 	mpz_shash(foo, 2, foo2, bar2);
 	assert(!mpz_cmp(foo, bar));
-	mpz_set_str(bar, "qxjmrvklcnU9I7llRZPj8v8CPv4laINbogTTGfEcYCY", TMCG_MPZ_IO_BASE);
+	mpz_set_str(bar, "caQFcUSBfXMGxXaZMMXp26QwN1vGcBYkhOz2DPgN6S8", TMCG_MPZ_IO_BASE);
 	s = "1234567890ABCDEFGHIJKLMNOPQRSTUVWXYZ";
 	mpz_shash(foo, s);
 	assert(!mpz_cmp(foo, bar));
